"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import copy
import warnings
import logging
from ..engine import ScopeStack
from ..reader import Reader
from ..containers import Container, ParsingError
from ..containers.layers import Placeholder
from ..model import Model, Trainer, Evaluator, EvaluationHook, OutputHook
from ..backend import Backend
from ..optimizer import Optimizer, Adam
from ..loss import Loss
from ..providers import Provider, BatchProvider, ShuffleProvider
from ..supplier import Supplier
from ..utils import merge_dict, mergetools
from ..loggers import Logger

logger = logging.getLogger(__name__)

###############################################################################
class Specification:
	""" Class for loading and parsing specification files.
	"""

	DEFAULT_OPTIMIZER = Adam
	DEFAULT_PROVIDER = BatchProvider

	###########################################################################
	def __init__(self, source, engine):
		""" Creates a new specification.

			# Arguments

			source: str or dict. If it is a string, it is interpretted as a
				filename to an on-disk specification file. Otherwise, it is
				interpretted as an already-loaded, but unparsed, specification
				file.
			engine: Engine instance. The templating engine to use in parsing.
		"""
		if isinstance(source, str):
			filename = os.path.expanduser(os.path.expandvars(source))
			if not os.path.isfile(filename):
				raise IOError('No such file found: {}. Path was expanded to: '
					'{}.'.format(source, filename))
			self.filename = filename
			self.data = self.parse_source(
				engine,
				source=filename,
				context='top-level'
			)
		else:
			self.filename = None
			self.data = dict(source)

		self.containers = None
		self.model = None
		self.backend = None
		self.engine = engine

	###########################################################################
	def parse(self):
		""" Parses the specification.
		"""

		logger.info('Parsing specification...')

		# These are reserved section names.
		# The first name in each tuple is the one that we should rely on
		# internally when we reference `self.data` in Specification.
		builtin = {
			'settings' : ('settings', ),
			'train' : ('train', 'training'),
			'validate' : ('validate', 'validation'),
			'test' : ('test', 'testing'),
			'evaluate' : ('evaluate', 'evaluation'),
			'model' : ('model', ),
			'loss' : ('loss', )
		}

		# The scope stack.
		stack = []

		# Add default entries.
		stack.append({
			'filename' : self.filename,
			'raw' : copy.deepcopy(self.data),
			'parsed' : self.data
		})

		# Parse the settings (backends, globals, hyperparameters, ...)
		self._parse_section(
			self.engine, builtin['settings'], stack, include_key=False)

		# Parse all of the "action" sections.
		self._parse_section(
			self.engine, builtin['train'], stack, include_key=True)
		self._parse_section(
			self.engine, builtin['validate'], stack, include_key=True)
		self._parse_section(
			self.engine, builtin['test'], stack, include_key=True)
		self._parse_section(
			self.engine, builtin['evaluate'], stack, include_key=True)

		# Parse the model.
		self.containers = self._parse_model(
			self.engine, builtin['model'], stack, required=True)

		# Parse the loss function.
		self._parse_section(
			self.engine, builtin['loss'], stack, include_key=True)

		# Check for unused keys.
		for key in self.data.keys():
			for section in builtin.values():
				if key in section:
					break
			else:
				warnings.warn('Unexpected section in specification: "{}". '
					'This section will be ignored.', SyntaxWarning)

	###########################################################################
	def apply_provider_knowledge(self, provider):
		""" Enables inference of tensor shapes by modifying the parsed
			containers given provider information.
		"""
		logger.debug('Applying provider-inferred shapes to input layers.')
		if provider.keys is None:
			logger.debug('No provider keys available. Cannot infer shapes.')
			return

		sources = dict(zip(provider.keys, provider.sources))
		for container in self.containers:
			for target in container.get_children(
					recursive=True, include_self=True
				):
				if not isinstance(target, Placeholder):
					continue

				logger.debug('Trying to infer shape for input "%s"',
					target.name)

				if target.name not in sources:
					logger.warning('Could not find a data source for model '
						'input "%s". Maybe you meant one of these: %s',
						target.name, ', '.join(provider.keys()))
					continue

				source = sources.pop(target.name)
				shape = source.shape()

				if target.shape is None:
					logger.debug('Inferred shape for input "%s": %s',
						target.name, shape)
					target.set_shape(shape)

				elif target.shape != shape:
					logger.warning('The input placeholder "%s" in the model '
						'has an explicit shape %s which disagrees with the '
						'shape of the corresponding data source %s.',
						target.name, target.shape, source.shape())

				else:
					logger.debug('Input "%s" already has a shape.',
						target.name)

	###########################################################################
	def get_model(self, provider=None):
		""" Returns the parsed Model instance.
		"""
		if self.model is None:
			if self.containers is None:
				raise ValueError('No such model available.')
			self.model = Model(
				backend=self.get_backend(),
				containers=self.containers
			)
			self.model.parse(self.engine)
			if provider is not None:
				self.model.apply_provider_knowledge(provider)
			self.model.build()
		return self.model

	###########################################################################
	def get_backend(self):
		""" Creates a new backend from the specification.

			# Return value

			Backend instance

			# Notes

			- If no "global" section exists, or if no "backend" section exists
			  in the "global" section, this will try to instantiate any
			  supported backend installed on the system.
			- This passes `self.data['global']['backend']` to the Backend
			  factory method, if such keys exist; otherwise, it passes None to
			  the factory method.
		"""
		if self.backend is None:
			self.backend = Backend.from_specification(
				self.data.get('settings', {}).get('backend')
			)
		return self.backend

	###########################################################################
	def get_provider(self, section):
		""" Creates the provider corresponding to a part of the specification.

			# Arguments

			section: str. The name of the section to load the provider for.

			# Return value

			If the section exists and has at least the "data" or "provider"
			section defined, then this returns a Provider instance. Otherwise,
			returns None.
		"""
		if section in self.data:
			section = self.data[section]
			if not any(k in section for k in ('data', 'provider')):
				return None
		else:
			return None

		supplier_list = section.get('data') or []
		if not isinstance(supplier_list, (list, tuple)):
			raise ValueError('"data" section should be a list.')
		suppliers = [Supplier.from_specification(x) for x in supplier_list]

		provider_spec = dict(section.get('provider') or {})
		if 'name' in provider_spec:
			provider = Provider.get_provider_by_name(provider_spec.pop('name'))
		else:
			provider = Specification.DEFAULT_PROVIDER

		# TODO: merge_dict is good for different columns, but we may need to
		#   concatenate columns with the same names. Need ConcatenateSource.
		return provider(
			sources=merge_dict(
				*[supplier.get_sources() for supplier in suppliers]
			),
			**provider_spec		# Everything remaining are parameters.
		)

	###########################################################################
	def get_training_function(self):
		""" Returns a function that will train the model.
		"""
		if 'train' not in self.data:
			raise ValueError('Cannot construct training function. There is a '
				'missing "train" section.')

		if 'log' in self.data['train']:
			log = Logger.from_specification(self.data['train']['log'])
		else:
			log = None

		epochs = self.data['train'].get('epochs')

		provider = self.get_provider('train')

		if 'validate' in self.data:
			validation = self.get_provider('validate')
			validation_weights = self.data['validate'].get('weights')
			if validation_weights is None:
				best_valid = None
			elif isinstance(validation_weights, str):
				best_valid = validation_weights
			elif isinstance(validation_weights, dict):
				best_valid = validation_weights.get('best')
			else:
				raise ValueError('Unknown type for validation weights: {}'
					.format(validation_weights))
		else:
			validation = None
			best_valid = None

		train_weights = self.data['train'].get('weights')
		if train_weights is None:
			initial_weights = best_train = last_weights = None
		elif isinstance(train_weights, str):
			initial_weights = train_weights
			best_train = train_weights if best_valid is None else None
			last_weights = None
			initial_must_exist = False
		elif isinstance(train_weights, dict):
			initial_weights = train_weights.get('initial')
			best_train = train_weights.get('best')
			last_weights = train_weights.get('last')
			initial_must_exist = train_weights.get('must_exist', False)
		else:
			raise ValueError('Unknown weight specification for training: {}'
				.format(train_weights))

		expand = lambda x: os.path.expanduser(os.path.expandvars(x))
		initial_weights, best_train, best_valid, last_weights = [
			expand(x) if x is not None else x for x in
				(initial_weights, best_train, best_valid, last_weights)
		]

		model = self.get_model(provider)
		trainer = self.get_trainer()

		def func():
			""" Trains a model from a pre-packaged specification file.
			"""
			if initial_weights is not None:
				if os.path.exists(initial_weights):
					model.restore(initial_weights)
				elif initial_must_exist:
					logger.error('Configuration indicates that the weight '
						'file must exist, but the weight file was not found.')
					raise ValueError('Missing initial weight file. If you '
						'want to proceed anyway, set "must_exist" to "no" '
						'under the training "weights" section.')
				else:
					if log is not None and \
						(log.get_number_of_epochs() or 0) > 0:
						logger.warning('The initial weights are missing: %s. '
							'However, the logs suggest that this model has '
							'been trained previously. It is possible that the '
							'log file is corrupt/out-of-date, or that the '
							'weight files are not supposed to be missing.',
							initial_weights)
					else:
						logger.info('Ignoring missing initial weights: %s. If '
							'this is undesireable, set "must_exist" to "yes" '
							'in the approriate "weights" section.',
							initial_weights)
			return trainer.train(
				provider=provider,
				validation=validation,
				epochs=epochs,
				log=log,
				best_train=best_train,
				best_valid=best_valid,
				last_weights=last_weights
			)

		return func

	###########################################################################
	def get_testing_function(self):
		""" Returns a function that will test the model.
		"""
		if 'test' not in self.data:
			raise ValueError('Cannot construct testing function. There is a '
				'missing "test" section.')

		provider = self.get_provider('test')

		# No reason to shuffle things for this.
		if isinstance(provider, ShuffleProvider):
			provider.randomize = False

		weights = self.data['test'].get('weights')
		if weights is None:
			initial_weights = None
		elif isinstance(weights, str):
			initial_weights = weights
		elif isinstance(weights, dict):
			initial_weights = weights.get('initial')
		else:
			raise ValueError('Unknown weight specification for testing: {}'
				.format(weights))

		expand = lambda x: os.path.expanduser(os.path.expandvars(x))
		if initial_weights is not None:
			initial_weights = expand(initial_weights)

		model = self.get_model(provider)
		trainer = self.get_trainer(with_optimizer=False)

		def func():
			""" Tests a model from a pre-packaged specification file.
			"""
			if initial_weights is not None:
				if os.path.exists(initial_weights):
					model.restore(initial_weights)
				else:
					logger.error('No weight file found: %s. We will just '
						'proceed with default-initialized weights. This will '
						'test that the system works, but the results will be '
						'terrible.', initial_weights)
			else:
				logger.warning('No weight file specified. We will just '
					'proceed with default-initialized weights. This will '
					'test that the system works, but the results will be '
					'terrible.')
			return trainer.test(
				provider=provider,
				validating=False
			)

		return func

	###########################################################################
	def get_trainer(self, with_optimizer=True):
		""" Creates a new Trainer from the specification.

			# Return value

			Trainer instance.
		"""
		return Trainer(
			model=self.get_model(),
			loss=self.get_loss(),
			optimizer=self.get_optimizer() if with_optimizer else None
		)

	###########################################################################
	def get_optimizer(self):
		""" Creates a new Optimizer from the specification.

			# Return value

			Optimizer instance.
		"""
		if 'train' not in self.data:
			raise ValueError('Cannot construct optimizer. There is a missing '
				'"train" section.')

		spec = dict(self.data['train'].get('optimizer', {}))
		if 'name' in spec:
			optimizer = Optimizer.get_optimizer_by_name(spec.pop('name'))
		else:
			optimizer = Specification.DEFAULT_OPTIMIZER

		return optimizer(**spec)

	###########################################################################
	def get_loss(self):
		""" Creates a new Loss from the specification.

			# Return value

			If 'loss' is defined, then this returns a dictionary whose keys are
			output layer names and whose respective values are Loss instances.
			Otherwise, this returns None.
		"""
		if 'loss' not in self.data:
			return None

		spec = self.data['loss']
		if not isinstance(spec, (list, tuple)):
			raise ValueError('"loss" section expects a list of loss '
				'functions.')

		result = {}
		for entry in spec:
			entry = dict(entry)
			target = entry.pop('target', None)
			if target is None:
				raise ValueError('Each loss function in the "loss" list '
					'must have a "target" entry which names the output layer '
					'it is attached to.')

			name = entry.pop('name', None)
			if name is None:
				raise ValueError('Each loss function in the "loss" list '
					'must have a "name" entry which names the loss function '
					'to use for that output.')

			if target in result:
				logger.warning('A loss function for the "%s" output was '
					'already defined. We will replace it by this new one we '
					'just found.', target)
			result[target] = Loss.get_loss_by_name(name)(**entry)

		return result

	###########################################################################
	def get_evaluation_function(self):
		""" Returns a function that will evaluate the model.
		"""
		if 'evaluate' not in self.data:
			raise ValueError('Cannot construct evaluation function. There is '
				'a missing "evaluate" section.')

		provider = self.get_provider('evaluate')

		# No reason to shuffle things for this.
		if isinstance(provider, ShuffleProvider):
			provider.randomize = False

		weights = self.data['evaluate'].get('weights')
		if weights is None:
			initial_weights = None
		elif isinstance(weights, str):
			initial_weights = weights
		elif isinstance(weights, dict):
			initial_weights = weights.get('initial')
		else:
			raise ValueError('Unknown weight specification for evaluation: {}'
				.format(weights))

		destination = self.data['evaluate'].get('destination')
		if isinstance(destination, str):
			destination = OutputHook(path=destination)
		elif isinstance(destination, dict):
			destination = OutputHook(**destination)
		elif destination is not None:
			ValueError('Expected a string or dictionary value for '
				'"destination" in "evaluate". Received: {}'.format(
					destination))

		hooks = self.data['evaluate'].get('hooks', [])
		if not isinstance(hooks, (list, tuple)):
			raise ValueError('"hooks" should be a list of hook '
				'specifications.')
		hooks = [EvaluationHook.from_specification(spec) for spec in hooks]

		expand = lambda x: os.path.expanduser(os.path.expandvars(x))
		if initial_weights is not None:
			initial_weights = expand(initial_weights)

		model = self.get_model(provider)
		evaluator = self.get_evaluator()

		def func():
			""" Evaluates a model from a pre-packaged specification file.
			"""
			if initial_weights is not None:
				if os.path.exists(initial_weights):
					model.restore(initial_weights)
				else:
					logger.error('No weight file found: %s. We will just '
						'proceed with default-initialized weights. This will '
						'test that the system works, but the results will be '
						'terrible.', initial_weights)
			else:
				logger.warning('No weight file specified. We will just '
					'proceed with default-initialized weights. This will '
					'test that the system works, but the results will be '
					'terrible.')
			result, truth = evaluator.evaluate(
				provider=provider,
				callback=None
			)
			for hook in hooks:
				result = hook.apply(result, truth)
			if destination is not None:
				result = destination.apply(result, truth)
			return result

		return func

	###########################################################################
	def get_evaluator(self):
		""" Creates a new Evaluator from the specification.

			# Return value

			Evaluator instance.
		"""
		return Evaluator(
			model=self.get_model()
		)

	###########################################################################
	def _parse_model(self, engine, section, stack, *, required=True):
		""" Parses the top-level "model" entry.

			# Arguments

			engine: Engine instance. The templating engine to use in
				evaluation.
			section: str or list of str. The key indicating which top-level
				entry to parse. If a list, the first matching key is used; this
				allows for multiple aliases to the same section.
			stack: list. The list of scope dictionaries to use in evaluation.
			required: bool (default: True). If True, an exception is raised if
				the key is not present in the specification.

			# Return value

			If the model section is found, the list of containers is returned.
			If the model section is not found but is required, an exception is
			raised. If the model section is not found and isn't required, None
			is returned.
		"""
		if isinstance(section, str):
			section = (section, )

		key = None		# Not required, but shuts pylint up.
		for key in section:
			if key in self.data:
				break
		else:
			if required:
				raise ValueError(
					'Missing required section: {}'.format(', '.join(section)))
			else:
				return None

		if not isinstance(self.data[key], (list, tuple)):
			raise ValueError(
				'Section "{}" should contain a list of layers.'.format(key))

		containers = [
			Container.create_container_from_data(entry)
			for entry in self.data[key]
		]

		with ScopeStack(engine, stack):
			for container in containers:
				container.parse(engine)

		return containers

	###########################################################################
	def parse_source(self, engine, *,
		source=None, context=None, loaded=None):
		""" Parses a source, and its includes (recursively), and returns the
			merged sources.
		"""
		def get_id(filename):
			""" Returns a tuple which uniquely identifies a file.

				This function follows symlinks.
			"""
			filename = os.path.expanduser(os.path.expandvars(filename))
			stat = os.stat(filename)
			return (stat.st_dev, stat.st_ino)

		# Process the "loaded" iterable.
		loaded = loaded or set()

		if isinstance(source, str):
			filename = source
			strategy = None
		elif isinstance(source, dict):
			if 'source' not in source:
				raise ParsingError('Error while parsing source file {}. '
					'Missing required "source" key in the include '
					'dictionary: {}.'.format(context, source))
			filename = source.pop('source')
			strategy = source.pop('method', None)
			for k in source:
				warnings.warn('Warning while parsing source file {}. '
					'Ignoring extra key in an "include" dictionary: {}.'
					.format(context, k))
		else:
			raise ParsingError('Error while parsing source file {}. '
				'Expected each "include" to be a string or a dictionary. '
				'Received: {}'.format(context, source))

		logger.info('Parsing source: %s, included by %s.', source, context)

		expanded = os.path.expanduser(os.path.expandvars(filename))
		if not os.path.isfile(expanded):
			raise IOError('Error while parsing source file {}. No such '
				'source file found: {}. Path was expanded to: {}'.format(
				context, filename, expanded))
		file_id = get_id(expanded)
		if file_id in loaded:
			logger.warning('Skipping an already included source file: %s. '
				'This may have unintended consequences, since the merge '
				'order may result in different final results. You should '
				'refactor your specification to avoid circular includes.',
				expanded)
			return {}
		else:
			loaded.add(file_id)

		data = Reader.read_file(expanded)

		new_sources = data.pop('include', [])
		engine.evaluate(new_sources, recursive=True)
		if isinstance(new_sources, str):
			new_sources = [new_sources]
		elif not isinstance(new_sources, (list, tuple)):
			raise ValueError('Error while parsing file: {}. All "include" '
				'sections must be a single source file or a list of source '
				'files. Instead we received: {}'.format(
					expanded, new_sources))

		for x in new_sources:
			data = mergetools.deep_merge(
				self.parse_source(
					engine,
					source=x,
					context=expanded,
					loaded=loaded
				),
				data,
				strategy=strategy
			)

		return data

	###########################################################################
	def _parse_section(self, engine, section, stack, *,
		required=False, include_key=True):
		""" Parses a single top-level entry in the specification.

			# Arguments

			engine: Engine instance. The templating engine to use in
				evaluation.
			section: str or list of str. The key indicating which top-level
				entry to parse. If a list, the first matching key is used; this
				allows for multiple aliases to the same section.
			stack: list. The list of scope dictionaries to use in evaluation.
			required: bool (default: False). If True, an exception is raised if
				the key is not present in the specification.
			include_key: bool (default: False). If True, the parsed section is
				added to the scope stack as a dictionary with a single item,
				whose key is the name of the parsed section, and whose value is
				the evaluated section. If False, then the evaluated section is
				added directly to the scope stack (it must evaluate to a
				dictionary for this to work).

			# Return value

			The parsed section.

			# Note

			- This will replace the values of `self.data` with the evaluated
			  version.
		"""
		if isinstance(section, str):
			section = (section, )

		logger.debug('Parsing specification section: %s', section[0])

		key = None		# Not required, but shuts pylint up.
		for key in section:
			if key in self.data:
				break
		else:
			if required:
				raise ValueError(
					'Missing required section: {}'.format(', '.join(section)))
			else:
				return None

		with ScopeStack(engine, stack):
			evaluated = engine.evaluate(self.data[key], recursive=True)

		if include_key:
			stack.append({key : evaluated})
		else:
			if evaluated is not None:
				if not isinstance(evaluated, dict):
					raise ValueError(
						'Section "{}" should contain key/value pairs.'
						.format(key))
				stack.append(evaluated)

		# Now, just in case this was aliased, let's make sure we keep our
		# naming scheme consistent.
		for key in section:
			self.data[key] = evaluated

		return evaluated

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import gzip
import logging
import struct
import tempfile
import os
import hashlib

import requests
import kur
import numpy

logger = logging.getLogger(__name__)

################################################################################
def download_data(url, shasum):
	""" Downloads a URL to the system temporary directory.

		# Arguments

		url: str. The URL of the resource to download.
		shasum: str. The SHA256 hash of the resource content.

		# Return value

		The path of the file on the local system.

		# Notes

		This will only download the file if it doesn't already exist or if its
		checksum fails.
	"""

	_, filename = os.path.split(url)
	target = os.path.join(tempfile.gettempdir(), filename)

	if os.path.isfile(target):
		with open(target, 'rb') as fh:
			data = fh.read()
		sha = hashlib.sha256()
		sha.update(data)
		if sha.hexdigest() == shasum:
			print('Data present: {}'.format(target))
			return target

		print('Corrupt file present. Re-downloading: {}'.format(url))
	else:
		print('Downloading: {}'.format(url))

	data = requests.get(url)
	sha = hashlib.sha256()
	sha.update(data.content)
	if sha.hexdigest() != shasum:
		raise ValueError('Failed integrity check.')

	with open(target, 'wb') as fh:
		fh.write(data.content)

	return target

################################################################################
def load_labels(filename):
	""" Loads MNIST labels.

		# Arguments

		filename: str. The path to the file with the MNIST labels.

		# Return value

		A numpy array of one-hot labels.
	"""
	logger.debug('Loading MNIST labels: %s', filename)
	with gzip.open(filename) as fh:
		data = fh.read()

	magic, num_items = struct.unpack('>2I', data[0:8])
	if magic != 0x00000801:
		raise ValueError('Bad magic number.')

	logging.debug('Num items: %d', num_items)

	data = data[8:]
	if len(data) != num_items:
		logger.error('Expected %d items, got %d items', num_items, len(data))
		raise ValueError('Length mis-match.')

	data = numpy.fromstring(data, dtype='>u1').astype(numpy.int32)
	onehot = numpy.zeros((data.shape[0], 10))
	for i, row in enumerate(data):
		onehot[i][row] = 1
	return onehot

################################################################################
def load_images(filename):
	""" Loads MNST images.

		# Arguments

		filename: str. The path to the file containing the MNIST images.

		# Return value

		A numpy array of the images. The images will be mean-subtracted and
		normalized.
	"""
	logger.debug('Loading MNIST images: %s', filename)
	with gzip.open(filename) as fh:
		data = fh.read()

	magic, num_items, num_rows, num_cols = struct.unpack('>4I', data[0:16])
	if magic != 0x00000803:
		raise ValueError('Bad magic number.')

	logging.debug('Num items: %d', num_items)

	image_size = num_rows * num_cols
	data = data[16:]
	if len(data) != num_items * image_size:
		logger.error('Expected %d items, got %d items', num_items, len(data))
		raise ValueError('Length mis-match.')

	data = numpy.fromstring(data, dtype='>u1').astype(numpy.float32)
	data /= 256
	data -= data.mean()
	return data.reshape(num_items, num_rows, num_cols)

################################################################################
def get_training_data():
	""" Returns the training set.

		# Return value

		Dictionary with keys 'images' and 'labels' and values that are the
		MNIST normalized images and one-hot labels, respectively.
	"""
	images = load_images(download_data(
		'http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz',
		'440fcabf73cc546fa21475e81ea370265605f56be210a4024d2ca8f203523609'
	))

	labels = load_labels(download_data(
		'http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz',
		'3552534a0a558bbed6aed32b30c495cca23d567ec52cac8be1a0730e8010255c'
	))

	return {'images' : images, 'labels' : labels}

################################################################################
def get_testing_data():
	""" Returns the testing set.

		# Return value

		Dictionary with keys 'images' and 'labels' and values that are the
		MNIST normalized images and one-hot labels, respectively.
	"""
	images = load_images(download_data(
		'http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz',
		'8d422c7b0a1c1c79245a5bcf07fe86e33eeafee792b84584aec276f5a2dbc4e6'
	))

	labels = load_labels(download_data(
		'http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz',
		'f7ae60f92e00ec6debd23a6088c31dbd2371eca3ffa0defaefb259924204aec6'
	))

	return {'images' : images, 'labels' : labels}

################################################################################
def main():									# pylint: disable=too-many-locals
	""" Training and evaluation example using the MNIST handwriting datset.

		# References

		- http://yann.lecun.com/exdb/mnist/
	"""

	# How much the of the data set to load.
	# 1 = all of it (100%)
	# 0.01 = 1% of it.
	# This is mostly useful on slower machines where it is more important to see
	# how the example works than to get good results.
	ratio = 0.01
	print('Using {:.0f}% of the data.'.format(ratio*100))

	# Where to save/load the model weights.
	weight_file = os.path.join(tempfile.gettempdir(), 'mnist.weights')

	# Load the specification.
	specification = kur.reader.Reader.read_file('mnist.yml')

	# Get the training data.
	data = get_training_data()

	# 50,000 training + 10,000 validation
	#     = 60,000 total images in the training set.
	training = {k : v[:int(50000*ratio)] for k, v in data.items()}
	validation = {k : v[-int(10000*ratio):] for k, v in data.items()}

	training_provider = kur.providers.BatchProvider(
		sources={
			'images' : kur.sources.VanillaSource(training['images']),
			'labels' : kur.sources.VanillaSource(training['labels'])
		},
		batch_size=32
	)

	validation_provider = kur.providers.BatchProvider(
		sources={
			'images' : kur.sources.VanillaSource(validation['images']),
			'labels' : kur.sources.VanillaSource(validation['labels'])
		},
		batch_size=32
	)

	# Create the Kur model.
	model = kur.model.Model(
		backend=kur.backend.Backend.get_any_supported_backend()(),
		containers=[kur.containers.Container.create_container_from_data(entry)
			for entry in specification['model']]
	)

	# Parse and build the model.
	engine = kur.engine.JinjaEngine()
	model.parse(engine)
	model.apply_provider_knowledge(training_provider)
	model.build()

	# Create the training object.
	trainer = kur.model.Trainer(
		model=model,
		loss=kur.loss.Loss.get_loss_by_name('categorical_crossentropy')(),
		optimizer=kur.optimizer.Optimizer.get_optimizer_by_name('adam')()
	)

	if os.path.isfile(weight_file):
		print('Loading weights from:', weight_file)
		model.restore(weight_file)

	# Train the model.
	trainer.train(
		provider=training_provider,
		validation=validation_provider,
		epochs=1,
		log=None
	)

	print('Saving weights to:', weight_file)
	model.save(weight_file)

	# Create the evaluation object.
	evaluator = kur.model.Evaluator(model=model)

	# Get the testing data.
	data = get_testing_data()

	data = {k : v[:int(10000*ratio)] for k, v in data.items()}

	# Evaluate the model.
	evaluated, _ = evaluator.evaluate(
		provider=kur.providers.BatchProvider(
			sources={
				'images' : kur.sources.VanillaSource(data['images'])
			},
			batch_size=32,
			randomize=False
		)
	)

	# Print out statistics.
	for k, v in evaluated.items():
		good = 0
		total = 0

		# Compare the predictions to the labels.
		for result in kur.providers.BatchProvider(
			sources={
				'labels' : kur.sources.VanillaSource(data['labels']),
				'evaluated' : kur.sources.VanillaSource(v)
			},
			randomize=False,
			batch_size=32
		):
			truth = result['labels'].argmax(axis=1)
			predict = result['evaluated'].argmax(axis=1)
			bad = numpy.count_nonzero(truth - predict)
			good += len(truth) - bad
			total += len(truth)

		print('{}: {}/{} correct'.format(k, good, total))

################################################################################
if __name__ == '__main__':
	main()

#### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ast
import jinja2
from .engine import Engine

###############################################################################
class JinjaEngine(Engine):
	""" An evaluation engine which uses Jinja2 for templating.
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new Jinja2 templating engine.
		"""

		# Call the parent
		super().__init__(*args, **kwargs)

		# Create a Jinja2 environment. We could use jinja2.Template().render()
		# directly, but having an environment gives us more control over, e.g.,
		# custom filters.
		self.env = jinja2.Environment()

		# Registering custom filters is described here:
		#   http://jinja.pocoo.org/docs/dev/api/#custom-filters

		# Built-in Jinja2 filters are listed here:
		#   http://jinja.pocoo.org/docs/dev/templates/#builtin-filters

	###########################################################################
	def _evaluate(self, expression):
		""" Evaluates an expression in the current scope.

			# Arguments

			expression: str. The string to evaluate.

			# Return value

			The evaluated expression (some Python object/class).
		"""
		result = self.env.from_string(expression).render(**self._scope)

		# Jinja2's `render()` will return a string which is a valid Python
		# expression (e.g., passing it through `eval` will succeed). However,
		# if you reference, e.g., a list that Jinja renders, the list will get
		# printed as a string. So we use `ast.literal_eval()` to turn it back
		# into a Python object. This may have unintended consequences, such as
		# turning the literal string "None" into the `None` Python object.
		# But it's better than nothing.
		try:
			result = ast.literal_eval(result)
		except (ValueError, SyntaxError):
			pass

		return result

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer
from . import ParsingError

###############################################################################
class Reuse(Layer):						# pylint: disable=too-few-public-methods
	""" Reuse weights/operations from another container, rather than creating
		new tensor operations. This allows for weight-sharing between
		containers.

		# Example

		```
		reuse:
		  target: NAME
		```
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Create a new reuse container.
		"""
		super().__init__(*args, **kwargs)
		self.target = None
		self.reentrant = False

	###########################################################################
	def _parse(self, engine):
		""" Parse the debug statement and print it.
		"""
		super()._parse(engine)

		if 'target' not in self.args:
			raise ParsingError('Missing key "target" in Reuse layer.')
		self.target = engine.evaluate(self.args['target'])

	###########################################################################
	def _build(self, backend):
		""" Return a reference to another layer.
		"""
		if self.reentrant:
			raise ParsingError('Circular reference found while build reuse '
				'layer: {}'.format(self.name))

		target = self.get_relative_by_name(self.target)
		if target is None:
			raise ParsingError(
				'No such reference found for reuse: {}'.format(self.target))

		if target._built is None:			# pylint: disable=protected-access
			self.reentrant = True
			target.build(backend)
			self.reentrant = False

		yield from target.build(backend)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import OrderedDict

###############################################################################
def deep_merge(*args, strategy=None):
	""" Merges data structures together.

		# Arguments

		args: tuple. The list of data structures that should be merged.
		strategy: str or None (default: None). The strategy to use in merging
			the data structures. If None, it defaults to "blend". See
			`Strategies` below.

		# Strategies

		- merge: dictionaries are merged recursively. This means that if a key
		  is not present in both dictionaries, then the key and its value are
		  included in the merged result. If the key is in both dictionaries,
		  its value is recursively merged if both values are themselves
		  dictionaries, but only the last value is kept if the types are
		  different.
		- blend: dictionaries are merged recursively (as with "merge").
		  List/tuples are also merged in a similar manner, elementwise.
		- concat: dictionaries are merged recursively (as with "merge").
		  List/tuples are concatenated.
	"""

	func = {
		'merge' : _merge,
		'blend' : _blend,
		'concat' : _concat
	}.get(strategy or 'blend')
	if func is None:
		raise ValueError('Invalid strategy: {}'.format(strategy))

	if not args:
		return args
	elif len(args) == 1:
		return args[0]
	else:
		result = args[0]
		for x in args[1:]:
			result = func(result, x)
		return result

###############################################################################
def _blend(x, y):								# pylint: disable=invalid-name
	""" Implements the "blend" strategy for `deep_merge`.
	"""
	if isinstance(x, (dict, OrderedDict)):
		if not isinstance(y, (dict, OrderedDict)):
			return y
		return _merge(x, y, recursion_func=_blend)

	if isinstance(x, (list, tuple)):
		if not isinstance(y, (list, tuple)):
			return y
		result = [_blend(*i) for i in zip(x, y)]
		if len(x) > len(y):
			result += x[len(y):]
		elif len(x) < len(y):
			result += y[len(x):]
		return result

	return y

###############################################################################
def _concat(x, y):								# pylint: disable=invalid-name
	""" Implements the "concat" strategy for `deep_merge`.
	"""
	if isinstance(x, (dict, OrderedDict)):
		if not isinstance(y, (dict, OrderedDict)):
			return y
		return _merge(x, y, recursion_func=_concat)

	if isinstance(x, (list, tuple)):
		if not isinstance(y, (list, tuple)):
			return y
		return [i for L in (x, y) for i in L]

	return y

###############################################################################
def _merge(x, y, recursion_func=None):			# pylint: disable=invalid-name
	""" Implements the "merge" strategy for `deep_merge`.
	"""
	recursion_func = recursion_func or _merge

	if not any(isinstance(i, (dict, OrderedDict)) for i in (x, y)):
		return y

	result = {}

	for k, v in x.items():
		if k in y:
			result[k] = recursion_func(v, y[k])
		else:
			result[k] = v

	for k, v in y.items():
		if k not in x:
			result[k] = v

	return result

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Convolution(Layer):				# pylint: disable=too-few-public-methods
	""" A vanilla convolution layer.

		# Properties

		kernels: int (required). The number of kernels (also called filters) to
			use.
		size: int or list of ints (required). The size of each kernel. If an
			integer is used, it is interpretted as a one-dimensional
			convolution, the same as if it were put into a length-1 list.
		strides: int or list of ints (optional; default: 1 in each dimension).
			The stride (subsampling rate) between convolutions. If a list, it
			must be the same length as `size` and specify the stride in each
			respective dimension. If a single number, it is used as the stride
			in each dimension.
		activation: one of (relu) or None (optional; default; relu). The
			activation to apply after the convolution.

		# Example

		```
		convolution:
		  kernels: 64
		  size: [2, 2]
		  strides: 1
		  activation: relu
		```
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new expand layer.
		"""
		super().__init__(*args, **kwargs)
		self.kernels = None
		self.size = None
		self.strides = None
		self.activation = None

	###########################################################################
	def _parse(self, engine):
		""" Parses out the convolution layer.
		"""
		# Always call the parent.
		super()._parse(engine)

		# Parse self
		if 'kernels' not in self.args:
			raise ParsingError('Missing key "kernels" in convolution '
				'container.')
		self.kernels = engine.evaluate(self.args['kernels'])
		try:
			self.kernels = int(self.kernels)
		except ValueError:
			raise ParsingError('"kernels" must evaluate to an integer.')

		if 'size' not in self.args:
			raise ParsingError('Missing key "size" in convolution container.')
		self.size = engine.evaluate(self.args['size'], recursive=True)
		if not isinstance(self.size, (list, tuple)):
			self.size = [self.size]
		if not 1 <= len(self.size) <= 3:
			raise ParsingError('Only convolutions with dimensions 1, 2, or 3 '
				'are supported.')
		for i in range(len(self.size)):
			try:
				self.size[i] = int(self.size[i])
			except ValueError:
				raise ParsingError('All "size" entries must evaluate to '
					'integers. We received this instead: {}'
					.format(self.size[i]))

		if 'strides' in self.args:
			self.strides = engine.evaluate(self.args['strides'], recursive=True)
			if not isinstance(self.strides, (list, tuple)):
				try:
					self.strides = int(self.strides)
				except ValueError:
					raise ParsingError('"strides" must evaluate to an integer '
						'or a list of integers.')
				self.strides = [self.strides] * len(self.size)
			else:
				if len(self.strides) != len(self.size):
					raise ParsingError('If "strides" is a list, it must be the '
						'same length as "size".')
				for i in range(len(self.strides)):
					try:
						self.strides[i] = int(self.strides[i])
					except ValueError:
						raise ParsingError('Each element of "strides" must '
							'evaluate to an integer.')
		else:
			self.strides = [1] * len(self.size)

		if 'activation' in self.args:
			self.activation = engine.evaluate(self.args['activation'])
		else:
			self.activation = None

	###########################################################################
	def _build(self, backend):
		""" Instantiates the layer with the given backend.
		"""
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error

			kwargs = {
				'nb_filter' : self.kernels,
				'activation' : self.activation or 'linear',
				'border_mode' : 'same',
				'name' : self.name
			}

			if len(self.size) == 1:
				func = L.Convolution1D
				kwargs.update({
					'filter_length' : self.size[0],
					'subsample_length' : self.strides[0]
				})
			elif len(self.size) == 2:
				func = L.Convolution2D
				kwargs.update({
					'nb_col' : self.size[0],
					'nb_row' : self.size[1],
					'subsample' : self.strides
				})
			elif len(self.size) == 3:
				func = L.Convolution3D
				kwargs.update({
					'kernel_dim1' : self.size[0],
					'kernel_dim2' : self.size[1],
					'kernel_dim3' : self.size[2],
					'subsample' : self.strides
				})
			else:
				raise ValueError('Unhandled convolution dimension: {}. This is '
					'a bug.'.format(len(self.size)))

			yield func(**kwargs)

			if self.activation:
				yield L.Activation(self.activation)
		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import tqdm
import numpy
from ..utils import get_any_value

logger = logging.getLogger(__name__)

###############################################################################
class Evaluator:
	""" Class for using trained models to predict new outputs.
	"""

	###########################################################################
	def __init__(self, model):
		""" Creates a new evaluator.

			# Arguments

			model: Model instance. The model to evaluate.
		"""
		self.model = model
		self._compiled = None

	###########################################################################
	def compile(self, recompile=False):
		""" Compiles a model.

			This generates a backend-specific representation of the model,
			suitable for prediction.

			# Arguments

			recompile: bool (default: False). If the model has already been
				compiled, it is not compiled again unless this flag is True.

			# Return value

			None
		"""

		if self._compiled is not None and not recompile:
			return

		if not self.model.is_built():
			logger.warning('This model has never been built before. We are '
				'going to try to build it now. But the model should always be '
				'built with Model.build() before trying to compile it, just '
				'to ensure that everything has been parsed as you expect.')
			self.model.build()

		self._compiled = self.model.backend.compile(
			model=self.model
		)

	###########################################################################
	def evaluate(self, provider, callback=None):
		""" Evaluates the model on some data.

			# Arguments

			provider: Provider instance. The data provider which serves the
				data to be evaluated.
			callback: function or None. If not None, the callback is called
				after each evaluation batch and is passed two parameters:
				`predicted` and `truth`, where `predicted` is the model output
				and `truth` is the ground truth data (if provided by
				`provider`; otherwise, `truth` is set to `None`).

			# Return value

			If `callback` is None, then this returns a tuple `(predicted,
			truth)`, where `predicted` is a dictionary whose keys are the names
			of the output nodes of the model, and whose respective values are
			arrays of predictions (one row per input sample). If the provider
			provides ground truth information, then `truth` has a similar
			structure to `predicted`; if ground truth information is not
			available, then `truth` is None.

			Otherwise, if `callback` is not None, this returns None.
		"""

		self.compile()

		result = None
		truth = None
		has_truth = None
		total = len(provider)
		n_entries = 0

		with tqdm.tqdm(
					total=total,
					unit='samples',
					desc='Evaluating'
				) as pbar:

			for batch in provider:
				evaluated = self.model.backend.evaluate(
					model=self.model,
					data=batch,
					compiled=self._compiled
				)
				batch_size = len(get_any_value(batch))

				if has_truth is None:
					has_truth = all(k in batch for k in self.model.outputs)

				if callback is None:
					if total is None:
						if result is None:
							result = {k : [] for k in self.model.outputs}
						for k, v in evaluated.items():
							result[k].append(v)

						if has_truth:
							if truth is None:
								truth = {k : [] for k in self.model.outputs}
							for k in truth:
								truth[k].append(batch[k])
					else:
						if result is None:
							result = {
								k : numpy.empty(
									shape=(total, ) + v.shape[1:],
									dtype=v.dtype
								) for k, v in evaluated.items()
							}
						for k, v in evaluated.items():
							result[k][n_entries:(n_entries+batch_size)] = v[:]

						if has_truth:
							if truth is None:
								truth = {
									k : numpy.empty(
										shape=(total, ) + v.shape[1:],
										dtype=v.dtype
									) for k, v in evaluated.items()
								}
							for k in truth:
								truth[k][n_entries:(n_entries+batch_size)] = \
									batch[k][:].reshape(
										(-1,) + truth[k].shape[1:]
									)
				else:
					callback(evaluated, truth)

				n_entries += batch_size
				pbar.update(batch_size)

		if callback is not None:
			return

		if total is None:
			for k, v in result.items():
				result[k] = numpy.concatenate(v)
			for k, v in truth.items():
				truth[k] = numpy.concatenate(v)

		return result, truth

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from kur.utils import get_subclasses

###############################################################################
class EvaluationHook:					# pylint: disable=too-few-public-methods
	""" Base class for all evaluation hooks.

		Evaluation hooks are post-evaluation callbacks that can take a crack at
		modifying or reviewing the evaluation results.
	"""

	###########################################################################
	@staticmethod
	def from_specification(spec):
		""" Creates a new evaluation hook from a specification.

			# Arguments

			spec: str or dict. The specification object that was given.

			# Return value

			New EvaluationHook instance.
		"""
		if isinstance(spec, str):
			name = spec
			target = EvaluationHook.get_hook_by_name(spec)
			params = {}
			meta = {}
		elif isinstance(spec, dict):
			candidates = {}
			for key in spec:
				try:
					cls = EvaluationHook.get_hook_by_name(key)
					candidates[key] = cls
				except ValueError:
					pass

			if not candidates:
				raise ValueError('No valid evaluation hook found in '
					'specification: {}'.format(spec))
			elif len(candidates) > 1:
				raise ValueError('Too many possible evaluation hooks found in '
					'specification: {}'.format(spec))
			else:
				name, target = candidates.popitem()
				meta = dict(spec)
				meta.pop(name)
				params = spec[name] or {}

		else:
			raise ValueError('Expected the evaluation hooks specification to '
				'be a string or dictionary. We got this instead: {}'.format(
				spec))

		# At this point:
		# - name: the name of the evaluation hook, as specified in the spec.
		# - target: the class of the evaluation hook
		# - meta: dictionary of "other" keys in the spec that sit alongside the
		#   evaluation hook specification.
		# - params: parameters to be passed to the evaluation hook.

		return target(**params)

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the evaluation hook.

			# Return value

			A lower-case string unique to this evaluation hook.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_hooks():
		""" Returns an iterator to the names of all evaluation hooks.
		"""
		for cls in get_subclasses(EvaluationHook):
			yield cls

	###########################################################################
	@staticmethod
	def get_hook_by_name(name):
		""" Finds a evaluation hook class with the given name.
		"""
		name = name.lower()
		for cls in EvaluationHook.get_all_hooks():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such evaluation hook with name "{}"'.format(name))

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new evaluation hook.
		"""
		if args or kwargs:
			raise ValueError('One or more unexpected or unsupported arguments '
				'to the evaluation hook: {}'.format(
					', '.join(args + list(kwargs.keys()))
			))

	###########################################################################
	def apply(self, data, truth=None):
		""" Applies the hook to the data.

			# Arguments

			data: dict. Dictionary whose keys are the names of the output layers
				in the model, and whose values are a list of output tensors.
			truth: dict or None (default: None). If not None, then a dictionary
				with the same keys as `data` which contains ground truth data.

			# Return value

			A dictionary in the same format as `data` which can be passed on to
			the next hook.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Operator

###############################################################################
class ContainerGroup(Operator):			# pylint: disable=too-few-public-methods
	""" A pseudo-container for holding other containers.

		The primary purpose for this container is to provide a unified namespace
		for all containers. This way, even if the model specification is split
		over multiple Containers, this ContainerGroup will still allow
		containers to "find" other containers.
	"""

	###########################################################################
	def __init__(self, containers):
		""" Creates a new container group.

			# Arguments

			containers: list of Containers. The containers to add as children.

			# Notes

			None of the child containers should already have a parent; if any of
			them do, a ParsingError is raised.
		"""
		super().__init__(data=None)

		for child in containers:
			self.add_child(child)

	###########################################################################
	def _parse(self, engine):
		""" Parse all children.
		"""
		for child in self.children:
			child.parse(engine)

	###########################################################################
	def _build(self, backend):
		""" Build all children.
		"""
		for child in self.children:
			yield from child.build(backend)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import gzip
import os
import struct
import warnings
import numpy

###############################################################################
def _read(fh, num_bytes):
	""" Reads a number of bytes, throwing an exception if that many bytes isn't
		available.
	"""
	result = fh.read(num_bytes)
	if len(result) != num_bytes:
		raise IOError('Unexpected end-of-file.')
	return result

###############################################################################
def save(filename, data, append=False):
	""" Saves data to an IDX file.
	"""
	if not isinstance(data, numpy.ndarray):
		data = numpy.array(data)

	data_info = {
		'f' : (0x0D, '>f4'),
		'u' : (0x0C, '>i4'),
		'i' : (0x0C, '>i4')
	}
	data_type = data_info.get(data.dtype.kind, None)
	if not data_type:
		raise ValueError('Data type is not understood: {}'.format(data.dtype))

	needs_copy = any((
		data.dtype.itemsize != int(data_type[1][-1]),
		data.dtype.byteorder != data_type[1][0],
		data.dtype.kind != data_type[1][-2]
	))
	if needs_copy:
		data = data.astype(data_type[1])

	if os.path.isfile(filename) and append:

		# Verify the header and update it.
		with open(filename, 'r+b') as fh:
			magic = _read(fh, 4)
			if magic[0] != 0 or magic[1] != 0:
				raise IOError('Bad header in IDX file.')
			if magic[2] != data_type[0]:
				raise IOError('Data type mis-match.')
			if magic[3] != data.ndim:
				raise IOError('Data dimensionality mis-match.')

			mark = fh.tell()
			shape = tuple(
				struct.unpack('>I', _read(fh, 4))[0]
				for _ in range(data.ndim)
			)

			if shape[1:] != data.shape[1:]:
				raise IOError('Data shape mis-match.')

			fh.seek(mark)
			fh.write(struct.pack('>I', shape[0] + data.shape[0]))

	else:

		# Write the new header.
		with open(filename, 'wb') as fh:
			fh.write(bytes([0, 0, data_type[0], data.ndim]))
			for shape in data.shape:
				fh.write(struct.pack('>I', shape))

	# Write the data.
	with open(filename, 'ab') as fh:
		fh.write(data.tostring())

###############################################################################
def load(filename):
	""" Loads an IDX file.

		Reference for file format: http://yann.lecun.com/exdb/mnist/
	"""

	def opener(filename):
		""" Returns a file handle to a binary file, optionally gzipped.
		"""
		with open(filename, 'rb') as fh:
			magic = fh.read(2)

		# Test if the GZIP magic number is present.
		# Reference: http://www.gzip.org/zlib/rfc-gzip.html#file-format
		if magic == b'\x1f\x8b':
			return gzip.open
		return open

	with opener(filename)(filename, 'rb') as fh:
		magic = _read(fh, 4)
		if magic[0] != 0 or magic[1] != 0:
			raise IOError('Bad header in IDX file.')

		data_info = {
			0x08 : ('B', 1),
			0x09 : ('b', 1),
			0x0B : ('h', 2),
			0x0C : ('i', 4),
			0x0D : ('f', 4),
			0x0E : ('d', 8)
		}

		data_type = data_info.get(magic[2])
		if data_type is None:
			raise IOError(
				'Bad data type in IDX header: 0x{:02X}'.format(magic[2]))

		ndim = magic[3]
		shape = tuple(
			struct.unpack('>I', _read(fh, 4))[0] for _ in range(ndim)
		)

		num_entries = numpy.product(shape)

		entry_type, entry_size = data_type
		total_size = entry_size * num_entries

		data = _read(fh, total_size)

		try:
			_read(fh, 1)
		except IOError:
			pass
		else:
			warnings.warn('IDX file has additional data at the end of the '
				'file. Ignoring this data...')

	result = numpy.fromstring(data, dtype='>{}'.format(entry_type))
	result = result.reshape(shape)
	return result

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

###############################################################################
class Source:						# pylint: disable=too-few-public-methods
	""" Base class for all data sources.

		Data sources are responsible for producing data that will be presented
		to the model for training, validation, or prediction (evaluation).

		Each data source provides one of the input or output tensors that the
		network needs.
	"""

	###########################################################################
	def __init__(self):
		""" Create a new data source.
		"""
		pass

	###########################################################################
	def __iter__(self):
		""" Return an iterator to the data.

			The iterator should return a tensor of shape `(X, ) +
			self.shape()`, where `X` is the number of entries provided by this
			data source.
		"""
		raise NotImplementedError

	###########################################################################
	def __len__(self):
		""" Returns the total number of entries that this source can return, if
			known.

			# Return value

			If the total number of entries that this source can return is
			known, it should be returned. In the simplest case, this is the
			same as the number of training samples in an epoch. If this number
			is unknown or effectively infinite (because it is from some
			real-time data generator or is being streamed from some
			uncontrolled source, for example), then this should return None.
		"""
		raise NotImplementedError

	###########################################################################
	def shape(self):
		""" Return the shape of the tensor (excluding batch size) returned by
			this data source.

			# Return value

			A shape tuple (length of each dimension in the tensor).
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import numpy
from . import ShuffleProvider

logger = logging.getLogger(__name__)

###############################################################################
class BatchProvider(ShuffleProvider): # pylint: disable=too-few-public-methods
	""" A provider which tries to return the same number of samples at every
		batch.

		All batches except for the last batch are guaranteed to be constant.
		The last batch might be different if there are not enough samples
		available to populate a complete batch.
	"""

	DEFAULT_BATCH_SIZE = 32

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the provider.
		"""
		return 'batch_provider'

	###########################################################################
	def __init__(self, batch_size=None, num_batches=None, *args, **kwargs):
		""" Creates a new batch provider.

			# Arguments

			batch_size: int. The number of batches to return each iteration.
		"""
		super().__init__(*args, **kwargs)
		self.batch_size = batch_size or BatchProvider.DEFAULT_BATCH_SIZE
		logger.info('Batch size set to: %d', self.batch_size)

		self.num_batches = num_batches

		if self.num_batches is not None:
			logger.info('Maximum number of batches set to: %d',
				self.num_batches)
			if self.entries is not None:
				self.entries = min(
					self.entries,
					self.num_batches * self.batch_size
				)

	###########################################################################
	def __iter__(self):
		""" Retrieves the next batch of data.

			# Return value

			A tensor of shape `(X, ) + self.shape()`, where `X` is the number
			of entries provided by this data source.
		"""

		# Should always be the first call in __iter__
		super().pre_iter()

		iterators = [iter(source) for source in self.sources]
		queues = [next(it) for it in iterators]
		sentinel = object()
		proceed = True
		batches_produced = 0

		while iterators and proceed:

			result = [None for it in iterators]

			# Go over each data source.
			for i, it in enumerate(iterators):

				# Get enough data out of each one.
				while len(queues[i]) < self.batch_size:

					# Get the next batch. If there isn't any data left, flag
					# it.  After all, we may have collected at least some data
					# during the `while` loop, and we should return that data.
					x = next(it, sentinel)
					if x is sentinel:
						proceed = False
						break

					# Add the data to the queue.
					queues[i] = numpy.concatenate([queues[i], x])

				# Get the data ready.
				result[i] = queues[i][:self.batch_size]
				queues[i] = queues[i][self.batch_size:]

			lens = {len(q) for q in result}
			if len(lens) == 1:
				if lens.pop():
					yield self.wrap(result)

					batches_produced += 1
					if self.num_batches is not None:
						if batches_produced >= self.num_batches:
							break
				else:
					break
			else:
				raise ValueError('Unable to accumulate equal-length batches '
					'for each data source.')

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import contextlib
import sys

###############################################################################
def redirect_stderr(x):
	""" Redirects stderr to another file-like object.

		This is some compatibility code to support Python 3.4.
	"""
	if hasattr(contextlib, 'redirect_stderr'):
		result = contextlib.redirect_stderr
	else:
		@contextlib.contextmanager
		def result(x):
			""" Stand-in for Python 3.5's `redirect_stderr`.

				Notes: Non-reentrant, non-threadsafe
			"""
			old_stderr = sys.stderr
			sys.stderr = x
			yield
			sys.stder = old_stderr

	return result(x)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

###############################################################################
def get_subclasses(cls, recursive=True):
	""" Enumerates all subclasses of a given class.

		# Arguments

		cls: class. The class to enumerate subclasses for.
		recursive: bool (default: True). If True, recursively finds all
			sub-classes.

		# Return value

		A list of subclasses of `cls`.
	"""
	sub = cls.__subclasses__()
	if recursive:
		for cls in sub:
			sub.extend(get_subclasses(cls, recursive=True))
	return sub

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ..utils import get_subclasses, get_any_value

###############################################################################
class Supplier:
	""" Base class for all suppliers.

		Suppliers are essentially factories for Sources. A single data file may
		have multiple data columns available, and a Supplier is responsible for
		parsing out the individual Sources.
	"""

	###########################################################################
	@staticmethod
	def from_specification(spec):
		""" Creates a new Supplier from a specification.
		"""

		if not isinstance(spec, dict):
			raise ValueError('Each element of the "input" list must be a '
				'dictionary.')

		candidates = set(
			cls.get_name() for cls in Supplier.get_all_suppliers()
		) & set(spec.keys())

		if not candidates:
			raise ValueError('Missing the key naming the Supplier type from '
				'an element of the "input" list. Valid keys are: {}'.format(
				', '.join(Supplier.get_all_suppliers())
			))
		if len(candidates) > 1:
			raise ValueError('Ambiguous supplier type in an element of the '
				'"input" list. Exactly one of the following keys must be '
				'present: {}'.format(', '.join(Supplier.get_all_suppliers())))

		name = candidates.pop()
		params = spec[name]

		supplier_name = spec.get('name')

		# All other keys must be parsed out by this point.

		if isinstance(params, dict):
			result = Supplier.get_supplier_by_name(name)(
				name=supplier_name, **params)
		elif isinstance(params, str):
			result = Supplier.get_supplier_by_name(name)(params,
				name=supplier_name)
		elif isinstance(params, (list, tuple)):
			result = Supplier.get_supplier_by_name(name)(*params,
				name=supplier_name)
		else:
			raise ValueError('Expected the Supplier to be given a dictionary, '
				'list, tuple, or string for parameters. Instead, we received: '
				'{}'.format(params))

		return result

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the supplier.

			# Return value

			A lower-case string unique to this supplier.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_suppliers():
		""" Returns an iterator to the names of all suppliers.
		"""
		for cls in get_subclasses(Supplier):
			yield cls

	###########################################################################
	@staticmethod
	def get_supplier_by_name(name):
		""" Finds a supplier class with the given name.
		"""
		name = name.lower()
		for cls in Supplier.get_all_suppliers():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such supplier with name "{}"'.format(name))

	###########################################################################
	def __init__(self, name=None):
		""" Creates a new supplier.
		"""
		self.name = name

	###########################################################################
	def is_delegate(self):						# pylint: disable=no-self-use
		""" Returns True if this supplier supplies suppliers.
		"""
		return False

	###########################################################################
	def get_delegates(self):					# pylint: disable=no-self-use
		""" Returns a dictionary or list of suppliers that this supplier
			supplies.

			This is important for nesting suppliers, for applications such as
			splitting.

			# Notes

			- If the Supplier implementation doesn't have any sub-suppliers,
			  then this call should return an empty dictioanry or empty list.
			- If the Supplier implementation does have sub-suppliers, then the
			  `is_delegate()` should return True, but calling `get_sources()`
			  should raise an exception.
		"""
		return {}

	###########################################################################
	def get_sources(self, sources=None):
		""" Returns a dictionary or list of data sources, as provided by this
			supplier.

			# Arguments

			sources: int or str, or a list of int or str. The sources to
				return. Passing a bare integer or string is equivalent to
				passing a list containing that single item. Only sources with
				those indices (or keys) are returned. If sources is None, then
				all sources are returned.

			# Return value

			A dictionary or list of Source instances.
		"""
		raise NotImplementedError

	###########################################################################
	def get_source(self, source):
		""" Returns a single data source.

			# Arguments

			source: str or int. If this supplier produces a list of sources,
				then this is the index of the source (as an integer). If this
				supplier produces a dictionary of sources, then this is the key
				corresponding to the desired source (as a string).

			# Return value

			A Source instance.
		"""
		# By default, we'll just use our more general `get_sources()` call to
		# do all the work.
		return get_any_value(self.get_sources(sources=(source, )))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

###############################################################################
from __future__ import print_function
import sys

###############################################################################
def error_message(msg):
	""" Prints an error message and exits.
	"""
	line_width = 60
	format_spec = '{{: ^{width}}}'.format(width=line_width)
	lines = [
		'', '',
		'='*line_width, '',
		'ERROR', '',
		msg, ''
		'See our troubleshooting page to get started:', '',
		'https://kur.deepgram.com/troubleshooting.html#installation', '',
		'='*line_width, '',
		"Uh, oh. There was an error. Look up there ^^^^ and you'll be",
		'training awesome models in no time!'
	]
	for line in lines:
		print(format_spec.format(line), file=sys.stderr)
	sys.exit(1)

###############################################################################
if sys.version_info < (3, 4):
	error_message('Kur requires Python 3.4 or later.')

###############################################################################
# pylint: disable=wrong-import-position
import os
import subprocess
import warnings
from setuptools import setup, find_packages
# pylint: enable=wrong-import-position

################################################################################
def readme():
	""" Return the README text.
	"""
	with open('README.rst') as fh:
		return fh.read()

################################################################################
def get_version():
	""" Gets the current version of the package.
	"""
	version_py = os.path.join(os.path.dirname(__file__), 'kur', 'version.py')
	with open(version_py, 'r') as fh:
		for line in fh:
			if line.startswith('__version__'):
				return line.split('=')[-1].strip().replace('"', '')
	raise ValueError('Failed to parse version from: {}'.format(version_py))

################################################################################
setup(
	# Package information
	name='kur',
	version=get_version(),
	description='Descriptive deep learning',
	long_description=readme(),
	keywords='deep learning',
	classifiers=[
	],

	# Author information
    url='https://github.com/deepgram/kur',
	author='Adam Sypniewski',
	author_email='adam@deepgram.com',
	license='Apache Software License '
		'(http://www.apache.org/licenses/LICENSE-2.0)',

	# What is packaged here.
	packages=find_packages(),

	# What to include.
	package_data={
		'': ['*.txt', '*.rst', '*.md']
	},

	# Dependencies
	install_requires=[
		'pyyaml>=3.12',
		'jinja2>=2.8',
		'numpy>=1.11.2',
		'requests>=2.12.3',
		'tqdm>=4.10.0',

		# Keras - the default backend (with Theano)
		'keras>=1.1.2',
		'theano>=0.8.2'
	],
	dependency_links=[
	],

	# Testing
	test_suite='tests',
	tests_require=[
		'pytest',
		'tensorflow'
	],
	setup_requires=['pytest-runner'],

	entry_points={
		'console_scripts' : ['kur=kur.__main__:main']
	},

	zip_safe=False
)

#### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from . import Operator

logger = logging.getLogger(__name__)

###############################################################################
class Debug(Operator):					# pylint: disable=too-few-public-methods
	""" Simple debug message printer.

		# Example

		Simple debug string:
		```
		debug: Hello World
		```

		With variable substitution:
		```
		debug: "There are {{ layers|length }} layers so far."
		```
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Create a new debug container.
		"""
		super().__init__(*args, **kwargs)
		self.message = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the debug statement and print it.
		"""
		super()._parse(engine)
		logger.debug('Container %s says: %s', self.name, self.args)

	###########################################################################
	def _build(self, backend):	# pylint: disable=unused-argument,no-self-use
		""" Debug statements don't produce layers.
		"""
		return iter(())

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Source, Shuffleable

###############################################################################
class VanillaSource(Source, Shuffleable):
	""" A data source which is simple wrapper around an in-memory array.
	"""

	###########################################################################
	def __init__(self, data, chunk_size=None, *args, **kwargs):
		""" Create a new vanilla data provider.

			# Arguments

			data: numpy-like array. The data to provide.
			chunk_size: int (default: None). The maximum number of samples to
				return at each iteration, or None to return all samples.

			# Notes

			- This class will not "take ownership" of the dataset; that is, it
			  only holds a reference to the data. This means that you are free
			  to shuffle the underlying dataset (or even change the shape or
			  length of the dataset, although this isn't necessarily a good
			  idea).
		"""
		super().__init__(*args, **kwargs)

		self.data = data
		self.chunk_size = chunk_size

	###########################################################################
	def __len__(self):
		""" Returns the number of samples this source provides.
		"""
		return self.data.shape[0]

	###########################################################################
	def shape(self):
		""" Returns the shape of the individual data samples.
		"""
		return self.data.shape[1:]

	###########################################################################
	def shuffle(self, indices):
		""" Applies a permutation to the data.
		"""
		if len(indices) > len(self):
			raise ValueError('Shuffleable was asked to apply permutation, but '
				'the permutation is longer than the length of the data set.')
		self.data[:len(indices)] = self.data[indices]

	###########################################################################
	def __iter__(self):
		""" Returns the next chunk of data.
		"""

		start = 0
		num_entries = len(self)
		while start < num_entries:
			if self.chunk_size is None:
				end = num_entries
			else:
				end = min(num_entries, start + self.chunk_size)

			yield self.data[start:end]
			start = end

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import shutil
import tqdm
from ..utils import get_any_value, CriticalSection

logger = logging.getLogger(__name__)

###############################################################################
class Trainer:
	""" Class for training models.
	"""

	###########################################################################
	def __init__(self, model, loss, optimizer=None):
		""" Creates a new trainer.

			# Arguments

			model: Model instance. The model to train.
			loss: Loss instance. The loss function to use in training.
			optimizer: Optimizer instance. The optimizer to use in training.
		"""
		self.model = model
		self.loss = loss
		self.optimizer = optimizer

		self._compiled = None

	###########################################################################
	def compile(self, recompile=False):
		""" Compiles a model.

			This generates a backend-specific representation of the model,
			suitable for training.

			# Arguments

			recompile: bool (default: False). If the model has already been
				compiled, it is not compiled again unless this flag is True.

			# Return value

			None
		"""

		if self._compiled is not None and not recompile:
			return

		if not self.model.is_built():
			logger.warning('This model has never been built before. We are '
				'going to try to build it now. But the model should always be '
				'built with Model.build() before trying to compile it, just '
				'to ensure that everything has been parsed as you expect.')
			self.model.build()

		logger.debug('Recompiling the model.')
		self._compiled = self.model.backend.compile(
			model=self.model,
			loss=self.loss,
			optimizer=self.optimizer
		)

	###########################################################################
	@staticmethod
	def source_shapes(provider):
		""" Prints debug information about the sources in this provider.
		"""
		if logger.isEnabledFor(logging.DEBUG):
			for i, source in enumerate(provider.sources):
				if provider.keys is None:
					name = "anonymous"
				else:
					name = provider.keys[i]

				logger.debug('Data source "%s": entries=%s, shape=%s',
					name, len(source), source.shape())

	###########################################################################
	def test(self, provider, validating=False):
		""" Tests/validates the model on some data.

			# Arguments

			provider: Provider instance. The data provider which serves the
				data to be evaluated on.
			validating: bool (default: False). If False, the console output
				refers to this process as "testing"; otherwise, it is referred
				to as "validating."

			# Return value

			The average loss across the validation set.
		"""

		self.compile()

		if validating:
			desc = ('Validating', 'Validation')
		else:
			desc = ('Testing', 'Test')

		# Create progress bar
		test_loss = None
		n_entries = 0
		with tqdm.tqdm(
					total=len(provider),
					unit='samples',
					desc='{}, loss=N/A'.format(desc[0])
				) as pbar:

			# Present each batch to the network.
			for batch in provider:
				batch_loss = self.model.backend.test(
					model=self.model,
					data=batch,
					compiled=self._compiled
				)

				batch_size = len(get_any_value(batch))

				#batch_loss = loss if isinstance(loss, float) \
				#	else sum(loss.values())

				new_entries = n_entries + batch_size

				if test_loss is None:
					test_loss = batch_loss
				else:
					test_loss = {
						k : v * (n_entries / new_entries) + \
							batch_loss[k] * (batch_size / new_entries)
						for k, v in test_loss.items()
					}
				#avg_loss = avg_loss * (n_entries / new_entries) + \
				#	batch_loss * (batch_size / new_entries)
				n_entries = new_entries

				# Update the progress bar
				pbar.set_description('{}, loss={:.3f}'.format(
					desc[0],
					sum(test_loss.values())
				))
				pbar.update(batch_size)

		if not n_entries:
			logger.warning('No data provided to validation/testing system.')
			return None

		logger.info('%s loss: %.3f', desc[1], sum(test_loss.values()))

		return test_loss

	###########################################################################
	def train(self, *args, last_weights=None, log=None, **kwargs):
		""" Trains the model on some data.

			This is the public entry point for training. It wraps the business
			logic so that it can handle error conditions.
		"""

		try:
			result = self.wrapped_train(*args, log=log, **kwargs)
		except:
			logger.exception('Exception raised during training.')
			raise
		else:
			return result
		finally:
			if last_weights is not None:
				logger.info('Saving most recent weights: %s', last_weights)
				with CriticalSection():
					self.model.save(last_weights)
			if log is not None:
				log.flush()

	###########################################################################
	def wrapped_train(self, provider, *, validation=None, epochs=None,
		log=None, best_train=None, best_valid=None):
		""" Trains the model on some data.

			# Arguments

			provider: Provider instance. The data provider which serves the
				data to be trained on.
			validation: Provider instance or None (default: None). The data
				provider which serves validation data.
			epochs: int or None (default: None). The number of epochs to train
				for, or None to train forever.
			log: Log instance or None (default: None). The logger to save
				training statistics with.

			# Return value

			None
		"""

		Trainer.source_shapes(provider)

		self.compile()

		if log is None:
			logger.info('No log specified, so no historical loss information '
				'is available.')
			best_train_loss = best_valid_loss = None
		else:
			best_train_loss = log.get_best_training_loss()
			if best_train_loss is not None:
				logger.info('Best historical training loss: %.3f',
					best_train_loss)
			else:
				logger.info('No historical training loss available from logs.')

			best_valid_loss = log.get_best_validation_loss()
			if best_valid_loss is not None:
				logger.info('Best historical validation loss: %.3f',
					best_valid_loss)
			else:
				logger.info(
					'No historical validation loss available from logs.')

		# The name of the most recently saved weight file. If the weights
		# change, this should be reset to None. Otherwise, saving weights can
		# be as simple as copying the previously saved file.
		saved_recent = None

		epoch = -1
		while True:
			epoch += 1
			if epochs is not None and epoch >= epochs:
				break

			# We are about to modify the weights. Invalidate the name of the
			# last weight file.
			saved_recent = None

			print()

			# Create progress bar
			train_loss = None
			n_entries = 0
			with tqdm.tqdm(
						total=len(provider),
						unit='samples',
						desc='Epoch {}/{}, loss=N/A'.format(epoch+1, epochs)
					) as pbar:

				# Present each batch to the network.
				for batch in provider:
					# The loss averaged over this batch.
					batch_loss = self.model.backend.train(
						model=self.model,
						data=batch,
						compiled=self._compiled
					)

					if log is not None:
						log.log_batch(batch_loss, 'loss')

					# How many entries we just processed.
					batch_size = len(get_any_value(batch))

					# How many entries we've processed this epoch.
					new_entries = n_entries + batch_size

					# Average the per-batch loss across training.
					# This will give us our average "training loss".
					if train_loss is None:
						train_loss = batch_loss
					else:
						train_loss = {
							k : v * (n_entries / new_entries) + \
								batch_loss[k] * (batch_size / new_entries)
							for k, v in train_loss.items()
						}

					n_entries = new_entries

					# Update the progress bar with the current loss.
					# Note that `batch_loss` is, in some sense, just the
					# instantaneous training loss. `train_loss` is the average
					# loss across the entire training set so far.
					pbar.set_description('Epoch {}/{}, loss={:.3f}'.format(
						epoch+1, epochs, sum(train_loss.values())
					))
					pbar.update(batch_size)

			if not n_entries:
				logger.warning('No data provided to training loop. Trying to '
					'move on to the next epoch.')
				continue

			cur_train_loss = sum(train_loss.values())
			logger.info('Training loss: %.3f', cur_train_loss)

			if best_train is not None:
				if best_train_loss is None or cur_train_loss < best_train_loss:
					logger.info('Saving best historical training weights: %s',
						best_train)
					best_train_loss = cur_train_loss
					with CriticalSection():
						self.model.save(best_train)
					saved_recent = best_train

			if log is not None:
				log.log_training(train_loss, 'loss')

			if validation is not None:
				# Continue with a validation run.
				validation_loss = self.test(
					provider=validation,
					validating=True
				)
				if validation_loss is None:
					continue

				cur_validation_loss = sum(validation_loss.values())
				if best_valid is not None:
					if best_valid_loss is None or \
							cur_validation_loss < best_valid_loss:
						logger.info(
							'Saving best historical validation weights: %s',
							best_valid
						)
						best_valid_loss = cur_validation_loss
						if saved_recent is None:
							with CriticalSection():
								self.model.save(best_valid)
							saved_recent = best_valid
						else:
							logger.debug(
								'Copying weights from: %s',
								saved_recent
							)
							with CriticalSection():
								shutil.rmtree(best_valid)
								shutil.copytree(saved_recent, best_valid)

				if log is not None:
					log.log_validation(validation_loss, 'loss')

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import ChainMap
import warnings
import logging

logger = logging.getLogger(__name__)

###############################################################################
class ScopeStack:					# pylint: disable=too-few-public-methods
	""" Context management for Engine scopes.

		# Example

		```python
		engine = Engine(...)
		with ScopeStack(engine, {'key' : 'value'}):
			pass
		```

		Or use multiple scopes:
		```python
		engine = Engine(...)
		scope1 = {'key' : 'value'}
		scope2 = {'key2' : 'value2'}
		with ScopeStack(engine, (scope1, scope2)):
			pass
		```
	"""

	###########################################################################
	def __init__(self, engine, scope):
		""" Creates a new scope stack.

			# Arguments

			engine: Engine instance. The engine to manage scope for.
			scope: dictionary or list of dictionaries. The scope(s) to add to
				the engine's scope stack during context management.
		"""
		self.engine = engine
		if not isinstance(scope, (list, tuple)):
			scope = [scope]
		self.scope = scope

	###########################################################################
	def __enter__(self):
		""" Enter context management and add scopes to the engine.
		"""
		for scope in self.scope:
			self.engine.scope(**scope)
		return self.engine

	###########################################################################
	def __exit__(self, exc_type, exc_value, traceback):
		""" Leave context management and pop the scopes.
		"""
		for _ in self.scope:
			self.engine.scope_pop()

###############################################################################
class Engine:
	""" Base class for all template engines.
	"""

	###########################################################################
	def __init__(self):
		""" Creates a new engine

			# Arguments

			scope: dict or None. The default key/value scope to use, or None to
				begin with an empty scope.
		"""
		self.state = {
			'tags' : {},
			'oldest': {},
			'layers' : []
		}
		self._scope = ChainMap(self.state)

	###########################################################################
	def scope(self, **kwargs):
		""" Create an additional scope.

			# Arguments

			kwargs: dict. The key/value scope to augment the current scope
				with.

			# Return value

			Returns the current instance (self).

			# Notes

			- This is intended to be used with context management, like so:

				```python
				engine = Engine()
				with engine.scope(a='b', c='d'):
					engine.evaluate(...)
				```

			- It can also be used standalone to add context layers, like so:

				```python
				engine = Engine()
				engine.scope(a='b', c='d')
				```
		"""
		self._scope = self._scope.new_child(kwargs)
		return self

	###########################################################################
	def scope_pop(self):
		""" Removes the most recent scope.
		"""
		self._scope = self._scope.parents

	###########################################################################
	def __enter__(self):
		""" Enter context management.
		"""
		return self

	###########################################################################
	def __exit__(self, exc_type, exc_value, traceback):
		""" Exit context management, popping the most recent scope.
		"""
		self.scope_pop()

	###########################################################################
	def register(self, container):
		""" Adds information about a Container to the templating engine.

			# Arguments

			container: Container instance. The container to register.
		"""
		# Tags
		for tag in container.tags:
			self.state['tags'][tag] = container.name

		# First-names
		for oldest in container.oldest:
			if oldest not in self.state['oldest']:
				self.state['oldest'] = oldest

		# Layers themselves
		self.state['layers'].append(container.name)

	###########################################################################
	def evaluate(self, expression, recursive=False):
		""" Evaluates an expression in the current scope.

			# Arguments

			expression: object. The object to evaluate. If it is a string, then
				it should be evaluated for template substitution and returned.
				Otherwise, the behavior depends on `recursive`. If `recursive`
				is True, then container types(dict, list, tuple), are
				recursively evaluated, preserving the original Python structure
				as much as possible. If `recursive` is False, or if the
				expression in not a container type, the expression is returned
				unchanged.
			recursive: bool (default: False). If True, container types are
				evaluated recursively; otherwise, container types are returned
				unchanged.

			# Return value

			The evaluated expression (some Python object/class).
		"""
		if isinstance(expression, (str, bytes)):
			return self._evaluate(expression)
		elif isinstance(expression, (int, float, complex, type(None))):
			pass
		elif recursive:
			if isinstance(expression, dict):
				return {k : self.evaluate(v, recursive=recursive)
					for k, v in expression.items()}
			elif isinstance(expression, list):
				return [self.evaluate(x, recursive=recursive)
					for x in expression]
			elif isinstance(expression, tuple):
				return tuple(self.evaluate(x, recursive=recursive)
					for x in expression)
			else:
				warnings.warn('Unexpanded Python type {}: {}'.format(
					type(expression),
					expression
				), UserWarning)
		return expression

	###########################################################################
	def _evaluate(self, expression):
		""" Evaluates a string expression in the current scope.

			# Arguments

			expression: str. The string to evaluate.

			# Return value

			The evaluated expression (some Python object/class)
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import hashlib
import tempfile
import logging

import tqdm
import requests

logger = logging.getLogger(__name__)

###############################################################################
def get_hash(path):
	""" Returns the SHA256 hash of a file.

		# Arguments

		path: str. Path to the file to produce a hash for.

		# Return value

		64-character lowercase string of hexadecimal digits representing the
		32-byte SHA256 hash of the file content.
	"""
	with open(path, 'rb') as fh:
		data = fh.read()
	sha = hashlib.sha256()
	sha.update(data)
	return sha.hexdigest()

###############################################################################
def download_file(url, sha256=None, target_dir=None):
	""" Downloads a URL to the system temporary directory.

		# Arguments

		url: str. The URL of the resource to download.
		sha256: str or None (default: None). The SHA256 hash of the resource
			content, or None to skip verification.
		target_dir: str or None (default: None). The target directory to store
			the file in. If it is None, it defaults to the system temp
			directory.

		# Return value

		The path of the file on the local system.

		# Notes

		This will only download the file if it doesn't already exist or if its
		checksum fails. If the checksum fails after a download, an exception is
		raised.
	"""

	if target_dir is None:
		target_dir = tempfile.gettempdir()

	if sha256 is not None:
		sha256 = sha256.lower()

	_, filename = os.path.split(url)
	target = os.path.join(target_dir, filename)

	# If the file already exists locally, verify its contents.
	if os.path.isfile(target):
		if sha256 is not None:
			if get_hash(target) == sha256:
				logger.info('File already present (checksum passed): %s',
					target)
				return target

			logger.info('Corrupt file present. Re-downloading: %s', url)
		else:
			logger.info('File already present (skipping checksum): %s', target)
			return target
	else:
		logger.info('File does not exist. Downloading: %s', url)

	response = requests.get(url, stream=True)
	with open(target, 'wb') as fh:
		for chunk in tqdm.tqdm(
			response.iter_content(),
			total=int(response.headers['content-length']),
			unit='bytes',
			unit_scale=True,
			desc='Downloading'
		):
			fh.write(chunk)

	logger.info('File downloaded: %s', target)

	if sha256 is not None:
		if get_hash(target) != sha256:
			raise ValueError('Failed integrity check.')

	return target

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .model import Model, ExtensionState, Extension
from .train import Trainer
from .evaluate import Evaluator
from .evaluation_hooks import EvaluationHook, OutputHook

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .. import Container

###############################################################################
class Operator(Container):					# pylint: disable=abstract-method
	""" Base class for operators, which are containers that do not produce
		backend-specific layers, but rather defer to their children to do so.
	"""

	###########################################################################
	def terminal(self):
		""" The identifying feature of an Operator is that it is non-terminal,
			so we implement the base-class method here.
		"""
		return False

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time

###############################################################################
class Timer:
	""" A simple timing class for gathering performance metrics.
	"""

	###########################################################################
	def __init__(self, started=True):
		""" Creates a new Timer.
		"""
		self.duration = 0
		self.mark = 0
		self.started = False
		if started:
			self.restart()

	###########################################################################
	def resume(self):
		""" Starts a previously paused timer.

			This has no effect if the timer was not paused.
		"""
		if self.started:
			return
		self.started = True
		self.mark = self._clock()

	###########################################################################
	def reset(self):
		""" Reset the current timer (zeros the accumulated time) without
			changing its running state.
		"""
		self.duration = 0

	###########################################################################
	def restart(self):
		""" Restarts a timer (zeros the accumulated time and starts the clock
			from now.
		"""
		self.reset()
		self.started = True
		self.mark = self._clock()

	###########################################################################
	def pause(self):
		""" Pauses a running clock.

			This has no effect if the timer was already paused.
		"""
		if not self.started:
			return
		self.duration += self._clock() - self.mark
		self.started = False

	###########################################################################
	def _clock(self):							# pylint: disable=no-self-use
		""" Returns an internal counter used for timing.
		"""
		return time.perf_counter()

	###########################################################################
	def __call__(self):
		""" Convenience method for getting the current accumulated time.
		"""
		return self.get()

	###########################################################################
	def get(self):
		""" Gets the current accumulated time (has no effect on running state).
		"""
		if self.started:
			return self.duration + (self._clock() - self.mark)
		else:
			return self.duration

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF
